@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2002
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -43,7 +43,7 @@
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: db_overflow.c,v 11.33 2001/06/14 20:56:40 krinsky Exp $";
+static const char revid[] = "$Id: db_overflow.c,v 11.46 2002/08/08 03:57:48 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -53,9 +53,9 @@ static const char revid[] = "$Id: db_overflow.c,v 11.33 2001/06/14 20:56:40 krin
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "db_am.h"
-#include "db_verify.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_am.h"
+#include "dbinc/db_verify.h"
 
 /*
  * Big key/data code.
@@ -83,6 +83,7 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 	u_int32_t *bpsz;
 {
 	DB_ENV *dbenv;
+	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	db_indx_t bytes;
 	u_int32_t curoff, needed, start;
@@ -90,6 +91,7 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 	int ret;
 
 	dbenv = dbp->dbenv;
+	mpf = dbp->mpf;
 
 	/*
 	 * Check if the buffer is big enough; if it is not and we are
@@ -136,13 +138,12 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 	 */
 	dbt->size = needed;
 	for (curoff = 0, p = dbt->data; pgno != PGNO_INVALID && needed > 0;) {
-		if ((ret = memp_fget(dbp->mpf, &pgno, 0, &h)) != 0) {
-			(void)__db_pgerr(dbp, pgno);
+		if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0)
 			return (ret);
-		}
+
 		/* Check if we need any bytes from this page. */
 		if (curoff + OV_LEN(h) >= start) {
-			src = (u_int8_t *)h + P_OVERHEAD;
+			src = (u_int8_t *)h + P_OVERHEAD(dbp);
 			bytes = OV_LEN(h);
 			if (start > curoff) {
 				src += start - curoff;
@@ -156,7 +157,7 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 		}
 		curoff += OV_LEN(h);
 		pgno = h->next_pgno;
-		(void)memp_fput(dbp->mpf, h, 0);
+		(void)mpf->put(mpf, h, 0);
 	}
 	return (0);
 }
@@ -174,9 +175,10 @@ __db_poff(dbc, dbt, pgnop)
 	db_pgno_t *pgnop;
 {
 	DB *dbp;
-	PAGE *pagep, *lastp;
-	DB_LSN new_lsn, null_lsn;
 	DBT tmp_dbt;
+	DB_LSN new_lsn, null_lsn;
+	DB_MPOOLFILE *mpf;
+	PAGE *pagep, *lastp;
 	db_indx_t pagespace;
 	u_int32_t sz;
 	u_int8_t *p;
@@ -188,7 +190,8 @@ __db_poff(dbc, dbt, pgnop)
 	 * item.
 	 */
 	dbp = dbc->dbp;
-	pagespace = P_MAXSPACE(dbp->pgsize);
+	mpf = dbp->mpf;
+	pagespace = P_MAXSPACE(dbp, dbp->pgsize);
 
 	ret = 0;
 	lastp = NULL;
@@ -207,27 +210,28 @@ __db_poff(dbc, dbt, pgnop)
 		 * have a partial record.
 		 */
 		if ((ret = __db_new(dbc, P_OVERFLOW, &pagep)) != 0)
-			return (ret);
-		if (DB_LOGGING(dbc)) {
+			break;
+		if (DBC_LOGGING(dbc)) {
 			tmp_dbt.data = p;
 			tmp_dbt.size = pagespace;
 			ZERO_LSN(null_lsn);
-			if ((ret = __db_big_log(dbp->dbenv, dbc->txn,
-			    &new_lsn, 0, DB_ADD_BIG, dbp->log_fileid,
-			    PGNO(pagep), lastp ? PGNO(lastp) : PGNO_INVALID,
+			if ((ret = __db_big_log(dbp, dbc->txn,
+			    &new_lsn, 0, DB_ADD_BIG, PGNO(pagep),
+			    lastp ? PGNO(lastp) : PGNO_INVALID,
 			    PGNO_INVALID, &tmp_dbt, &LSN(pagep),
 			    lastp == NULL ? &null_lsn : &LSN(lastp),
 			    &null_lsn)) != 0) {
 				if (lastp != NULL)
-					(void)memp_fput(dbp->mpf,
+					(void)mpf->put(mpf,
 					    lastp, DB_MPOOL_DIRTY);
 				lastp = pagep;
 				break;
 			}
 		} else
 			LSN_NOT_LOGGED(new_lsn);
-		/* Move lsn onto page. */
-		if (lastp)
+
+		/* Move LSN onto page. */
+		if (lastp != NULL)
 			LSN(lastp) = new_lsn;
 		LSN(pagep) = new_lsn;
 
@@ -235,7 +239,7 @@ __db_poff(dbc, dbt, pgnop)
 		    PGNO(pagep), PGNO_INVALID, PGNO_INVALID, 0, P_OVERFLOW);
 		OV_LEN(pagep) = pagespace;
 		OV_REF(pagep) = 1;
-		memcpy((u_int8_t *)pagep + P_OVERHEAD, p, pagespace);
+		memcpy((u_int8_t *)pagep + P_OVERHEAD(dbp), p, pagespace);
 
 		/*
 		 * If this is the first entry, update the user's info.
@@ -247,12 +251,12 @@ __db_poff(dbc, dbt, pgnop)
 		else {
 			lastp->next_pgno = PGNO(pagep);
 			pagep->prev_pgno = PGNO(lastp);
-			(void)memp_fput(dbp->mpf, lastp, DB_MPOOL_DIRTY);
+			(void)mpf->put(mpf, lastp, DB_MPOOL_DIRTY);
 		}
 		lastp = pagep;
 	}
-	if ((t_ret =
-	    memp_fput(dbp->mpf, lastp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+	if (lastp != NULL &&
+	    (t_ret = mpf->put(mpf, lastp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
 		ret = t_ret;
 	return (ret);
 }
@@ -270,25 +274,29 @@ __db_ovref(dbc, pgno, adjust)
 	int32_t adjust;
 {
 	DB *dbp;
+	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	int ret;
 
 	dbp = dbc->dbp;
-	if ((ret = memp_fget(dbp->mpf, &pgno, 0, &h)) != 0) {
-		(void)__db_pgerr(dbp, pgno);
+	mpf = dbp->mpf;
+
+	if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0) {
+		__db_pgerr(dbp, pgno, ret);
 		return (ret);
 	}
 
-	if (DB_LOGGING(dbc)) {
-		if ((ret = __db_ovref_log(dbp->dbenv, dbc->txn,
-		    &LSN(h), 0, dbp->log_fileid, h->pgno, adjust,
-		    &LSN(h))) != 0)
+	if (DBC_LOGGING(dbc)) {
+		if ((ret = __db_ovref_log(dbp,
+		    dbc->txn, &LSN(h), 0, h->pgno, adjust, &LSN(h))) != 0) {
+			(void)mpf->put(mpf, h, 0);
 			return (ret);
+		}
 	} else
 		LSN_NOT_LOGGED(LSN(h));
 	OV_REF(h) += adjust;
 
-	(void)memp_fput(dbp->mpf, h, DB_MPOOL_DIRTY);
+	(void)mpf->put(mpf, h, DB_MPOOL_DIRTY);
 	return (0);
 }
 
@@ -306,13 +314,16 @@ __db_doff(dbc, pgno)
 	DB *dbp;
 	PAGE *pagep;
 	DB_LSN null_lsn;
+	DB_MPOOLFILE *mpf;
 	DBT tmp_dbt;
 	int ret;
 
 	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+
 	do {
-		if ((ret = memp_fget(dbp->mpf, &pgno, 0, &pagep)) != 0) {
-			(void)__db_pgerr(dbp, pgno);
+		if ((ret = mpf->get(mpf, &pgno, 0, &pagep)) != 0) {
+			__db_pgerr(dbp, pgno, ret);
 			return (ret);
 		}
 
@@ -322,20 +333,20 @@ __db_doff(dbc, pgno)
 		 * decrement the reference count and return.
 		 */
 		if (OV_REF(pagep) > 1) {
-			(void)memp_fput(dbp->mpf, pagep, 0);
+			(void)mpf->put(mpf, pagep, 0);
 			return (__db_ovref(dbc, pgno, -1));
 		}
 
-		if (DB_LOGGING(dbc)) {
-			tmp_dbt.data = (u_int8_t *)pagep + P_OVERHEAD;
+		if (DBC_LOGGING(dbc)) {
+			tmp_dbt.data = (u_int8_t *)pagep + P_OVERHEAD(dbp);
 			tmp_dbt.size = OV_LEN(pagep);
 			ZERO_LSN(null_lsn);
-			if ((ret = __db_big_log(dbp->dbenv, dbc->txn,
-			    &LSN(pagep), 0, DB_REM_BIG, dbp->log_fileid,
+			if ((ret = __db_big_log(dbp, dbc->txn,
+			    &LSN(pagep), 0, DB_REM_BIG,
 			    PGNO(pagep), PREV_PGNO(pagep),
 			    NEXT_PGNO(pagep), &tmp_dbt,
 			    &LSN(pagep), &null_lsn, &null_lsn)) != 0) {
-				(void)memp_fput(dbp->mpf, pagep, 0);
+				(void)mpf->put(mpf, pagep, 0);
 				return (ret);
 			}
 		} else
@@ -369,13 +380,16 @@ __db_moff(dbp, dbt, pgno, tlen, cmpfunc, cmpp)
 	u_int32_t tlen;
 	int (*cmpfunc) __P((DB *, const DBT *, const DBT *)), *cmpp;
 {
-	PAGE *pagep;
 	DBT local_dbt;
+	DB_MPOOLFILE *mpf;
+	PAGE *pagep;
 	void *buf;
 	u_int32_t bufsize, cmp_bytes, key_left;
 	u_int8_t *p1, *p2;
 	int ret;
 
+	mpf = dbp->mpf;
+
 	/*
 	 * If there is a user-specified comparison function, build a
 	 * contiguous copy of the key, and call it.
@@ -390,27 +404,27 @@ __db_moff(dbp, dbt, pgno, tlen, cmpfunc, cmpp)
 			return (ret);
 		/* Pass the key as the first argument */
 		*cmpp = cmpfunc(dbp, dbt, &local_dbt);
-		__os_free(dbp->dbenv, buf, bufsize);
+		__os_free(dbp->dbenv, buf);
 		return (0);
 	}
 
 	/* While there are both keys to compare. */
 	for (*cmpp = 0, p1 = dbt->data,
 	    key_left = dbt->size; key_left > 0 && pgno != PGNO_INVALID;) {
-		if ((ret = memp_fget(dbp->mpf, &pgno, 0, &pagep)) != 0)
+		if ((ret = mpf->get(mpf, &pgno, 0, &pagep)) != 0)
 			return (ret);
 
 		cmp_bytes = OV_LEN(pagep) < key_left ? OV_LEN(pagep) : key_left;
 		tlen -= cmp_bytes;
 		key_left -= cmp_bytes;
-		for (p2 =
-		    (u_int8_t *)pagep + P_OVERHEAD; cmp_bytes-- > 0; ++p1, ++p2)
+		for (p2 = (u_int8_t *)pagep + P_OVERHEAD(dbp);
+		    cmp_bytes-- > 0; ++p1, ++p2)
 			if (*p1 != *p2) {
 				*cmpp = (long)*p1 - (long)*p2;
 				break;
 			}
 		pgno = NEXT_PGNO(pagep);
-		if ((ret = memp_fput(dbp->mpf, pagep, 0)) != 0)
+		if ((ret = mpf->put(mpf, pagep, 0)) != 0)
 			return (ret);
 		if (*cmpp != 0)
 			return (0);
@@ -457,7 +471,7 @@ __db_vrfy_overflow(dbp, vdp, h, pgno, flags)
 	pip->refcount = OV_REF(h);
 	if (pip->refcount < 1) {
 		EPRINT((dbp->dbenv,
-		    "Overflow page %lu has zero reference count",
+		    "Page %lu: overflow page has zero reference count",
 		    (u_long)pgno));
 		isbad = 1;
 	}
@@ -512,7 +526,7 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 
 	if (pip->type != P_OVERFLOW) {
 		EPRINT((dbp->dbenv,
-		    "Overflow page %lu of invalid type",
+		    "Page %lu: overflow page of invalid type %lu",
 		    (u_long)pgno, (u_long)pip->type));
 		ret = DB_VERIFY_BAD;
 		goto err;		/* Unsafe to continue. */
@@ -521,7 +535,8 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 	prev = pip->prev_pgno;
 	if (prev != PGNO_INVALID) {
 		EPRINT((dbp->dbenv,
-		    "First overflow page %lu has a prev_pgno", (u_long)pgno));
+	    "Page %lu: first page in overflow chain has a prev_pgno %lu",
+		    (u_long)pgno, (u_long)prev));
 		isbad = 1;
 	}
 
@@ -560,7 +575,7 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 		 */
 		if ((u_int32_t)p > refcount) {
 			EPRINT((dbp->dbenv,
-			    "Page %lu encountered twice in overflow traversal",
+			    "Page %lu: encountered twice in overflow traversal",
 			    (u_long)pgno));
 			ret = DB_VERIFY_BAD;
 			goto err;
@@ -588,8 +603,8 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 		if (!IS_VALID_PGNO(next)) {
 			DB_ASSERT(0);
 			EPRINT((dbp->dbenv,
-			    "Overflow page %lu has bad next_pgno",
-			    (u_long)pgno));
+			    "Page %lu: bad next_pgno %lu on overflow page",
+			    (u_long)pgno, (u_long)next));
 			ret = DB_VERIFY_BAD;
 			goto err;
 		}
@@ -599,8 +614,9 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 			return (ret);
 		if (pip->prev_pgno != pgno) {
 			EPRINT((dbp->dbenv,
-			    "Overflow page %lu has bogus prev_pgno value",
-			    (u_long)next));
+		"Page %lu: bad prev_pgno %lu on overflow page (should be %lu)",
+			    (u_long)next, (u_long)pip->prev_pgno,
+			    (u_long)pgno));
 			isbad = 1;
 			/*
 			 * It's safe to continue because we have separate
@@ -614,7 +630,7 @@ __db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
 	if (tlen > 0) {
 		isbad = 1;
 		EPRINT((dbp->dbenv,
-		    "Overflow item incomplete on page %lu", (u_long)pgno));
+		    "Page %lu: overflow item incomplete", (u_long)pgno));
 	}
 
 err:	if ((t_ret =
@@ -640,14 +656,16 @@ __db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
 	void **buf;
 	u_int32_t flags;
 {
+	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	int ret, t_ret;
 	u_int32_t bytesgot, bytes;
 	u_int8_t *src, *dest;
 
+	mpf = dbp->mpf;
+	h = NULL;
 	ret = t_ret = 0;
 	bytesgot = bytes = 0;
-	h = NULL;
 
 	while ((pgno != PGNO_INVALID) && (IS_VALID_PGNO(pgno))) {
 		/*
@@ -657,7 +675,7 @@ __db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
 		if ((ret = __db_salvage_markdone(vdp, pgno)) != 0)
 			break;
 
-		if ((ret = memp_fget(dbp->mpf, &pgno, 0, &h)) != 0)
+		if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0)
 			break;
 
 		/*
@@ -669,11 +687,11 @@ __db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
 			break;
 		}
 
-		src = (u_int8_t *)h + P_OVERHEAD;
+		src = (u_int8_t *)h + P_OVERHEAD(dbp);
 		bytes = OV_LEN(h);
 
-		if (bytes + P_OVERHEAD > dbp->pgsize)
-			bytes = dbp->pgsize - P_OVERHEAD;
+		if (bytes + P_OVERHEAD(dbp) > dbp->pgsize)
+			bytes = dbp->pgsize - P_OVERHEAD(dbp);
 
 		if ((ret = __os_realloc(dbp->dbenv,
 		    bytesgot + bytes, buf)) != 0)
@@ -686,7 +704,7 @@ __db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
 
 		pgno = NEXT_PGNO(h);
 
-		if ((ret = memp_fput(dbp->mpf, h, 0)) != 0)
+		if ((ret = mpf->put(mpf, h, 0)) != 0)
 			break;
 		h = NULL;
 	}
@@ -701,7 +719,7 @@ __db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
 	}
 
 	/* If we broke out on error, don't leave pages pinned. */
-	if (h != NULL && (t_ret = memp_fput(dbp->mpf, h, 0)) != 0 && ret == 0)
+	if (h != NULL && (t_ret = mpf->put(mpf, h, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret);
