@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2002
  *      Sleepycat Software.  All rights reserved.
  *
- * $Id: DbLock.java,v 11.6 2001/05/16 13:09:28 bostic Exp $
+ * $Id: DbLock.java,v 11.8 2002/01/11 15:52:36 bostic Exp $
  */
 
 package com.sleepycat.db;
@@ -18,11 +18,6 @@ public class DbLock
     protected native void finalize()
          throws Throwable;
 
-    // methods
-    //
-    public native void put(DbEnv env)
-         throws DbException;
-
     // get/set methods
     //
 
