@@ -1,25 +1,25 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2002
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: os_spin.c,v 11.7 2001/01/25 18:23:00 bostic Exp $";
+static const char revid[] = "$Id: os_spin.c,v 11.11 2002/07/12 18:56:56 bostic Exp $";
 #endif /* not lint */
 
 #include "db_int.h"
-#include "os_jump.h"
 
 /*
  * __os_spin --
  *	Return the number of default spins before blocking.
  */
 int
-__os_spin()
+__os_spin(dbenv)
+	DB_ENV *dbenv;
 {
 	SYSTEM_INFO SystemInfo;
 
@@ -27,8 +27,8 @@ __os_spin()
 	 * If the application specified a value or we've already figured it
 	 * out, return it.
 	 */
-	if (DB_GLOBAL(db_tas_spins) != 0)
-		return (DB_GLOBAL(db_tas_spins));
+	if (dbenv->tas_spins != 0)
+		return (dbenv->tas_spins);
 
 	/* Get the number of processors */
 	GetSystemInfo(&SystemInfo);
@@ -38,10 +38,10 @@ __os_spin()
 	 * is a reasonable value.
 	 */
 	if (SystemInfo.dwNumberOfProcessors > 1)
-		 DB_GLOBAL(db_tas_spins) = 50 * SystemInfo.dwNumberOfProcessors;
+		 dbenv->tas_spins = 50 * SystemInfo.dwNumberOfProcessors;
 	else
-		 DB_GLOBAL(db_tas_spins) = 1;
-	return (DB_GLOBAL(db_tas_spins));
+		 dbenv->tas_spins = 1;
+	return (dbenv->tas_spins);
 }
 
 /*
@@ -53,7 +53,7 @@ __os_yield(dbenv, usecs)
 	DB_ENV *dbenv;
 	u_long usecs;
 {
-	if (__db_jump.j_yield != NULL && __db_jump.j_yield() == 0)
+	if (DB_GLOBAL(j_yield) != NULL && DB_GLOBAL(j_yield)() == 0)
 		return;
 	__os_sleep(dbenv, 0, usecs);
 }
